@@ -24,10 +24,11 @@
 // NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 // SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-use crate::events::quic::QuicFrame;
 use crate::events::EventData;
 use crate::events::EventImportance;
 use crate::events::EventType;
+use crate::events::Eventable;
+use crate::events::ExData;
 
 /// A helper object specialized for streaming JSON-serialized qlog to a
 /// [`Write`] trait.
@@ -36,21 +37,15 @@ use crate::events::EventType;
 /// provided `Trace`.
 ///
 /// Serialization is progressively driven by method calls; once log streaming
-/// is started, `event::Events` can be written using `add_event()`. Some
-/// events can contain an array of `QuicFrame`s, when writing such an event,
-/// the streamer enters a frame-serialization mode where frames are be
-/// progressively written using `add_frame()`. This mode is concluded using
-/// `finished_frames()`. While serializing frames, any attempts to log
-/// additional events are ignored.
+/// is started, `event::Events` can be written using `add_event()`.
 ///
 /// [`Write`]: https://doc.rust-lang.org/std/io/trait.Write.html
 use super::*;
 
-#[derive(PartialEq, Debug)]
+#[derive(PartialEq, Eq, Debug)]
 pub enum StreamerState {
     Initial,
     Ready,
-    WritingFrames,
     Finished,
 }
 
@@ -60,17 +55,18 @@ pub struct QlogStreamer {
     qlog: QlogSeq,
     state: StreamerState,
     log_level: EventImportance,
-    first_frame: bool,
 }
 
 impl QlogStreamer {
-    /// Creates a QlogStreamer object.
+    /// Creates a [QlogStreamer] object.
     ///
-    /// It owns a `Qlog` object that contains the provided `Trace` containing
-    /// `Events`.
+    /// It owns a [QlogSeq] object that contains the provided [TraceSeq]
+    /// containing [Event]s.
     ///
-    /// All serialization will be written to the provided `Write` using the
+    /// All serialization will be written to the provided [`Write`] using the
     /// JSON-SEQ format.
+    ///
+    /// [`Write`]: https://doc.rust-lang.org/std/io/trait.Write.html
     #[allow(clippy::too_many_arguments)]
     pub fn new(
         qlog_version: String, title: Option<String>, description: Option<String>,
@@ -93,15 +89,21 @@ impl QlogStreamer {
             qlog,
             state: StreamerState::Initial,
             log_level,
-            first_frame: false,
         }
     }
 
     /// Starts qlog streaming serialization.
     ///
-    /// This writes out the JSON-serialized form of all initial qlog information
-    /// `Event`s are separately appended using `add_event()` and
-    /// `add_event_with_instant()`.
+    /// This writes out the JSON-SEQ-serialized form of all initial qlog
+    /// information. [Event]s are separately appended using [add_event()],
+    /// [add_event_with_instant()], [add_event_now()],
+    /// [add_event_data_with_instant()], or [add_event_data_now()].
+    ///
+    /// [add_event()]: #method.add_event
+    /// [add_event_with_instant()]: #method.add_event_with_instant
+    /// [add_event_now()]: #method.add_event_now
+    /// [add_event_data_with_instant()]: #method.add_event_data_with_instant
+    /// [add_event_data_now()]: #method.add_event_data_now
     pub fn start_log(&mut self) -> Result<()> {
         if self.state != StreamerState::Initial {
             return Err(Error::Done);
@@ -119,7 +121,6 @@ impl QlogStreamer {
 
     /// Finishes qlog streaming serialization.
     ///
-    /// The JSON-serialized output has remaining close delimiters added.
     /// After this is called, no more serialization will occur.
     pub fn finish_log(&mut self) -> Result<()> {
         if self.state == StreamerState::Initial ||
@@ -135,32 +136,21 @@ impl QlogStreamer {
         Ok(())
     }
 
-    /// Writes a JSON-serialized `Event` using `std::time::Instant::now()`.
-    ///
-    /// Some qlog events can contain `QuicFrames`. If this is detected `true` is
-    /// returned and the streamer enters a frame-serialization mode that is only
-    /// concluded by `finish_frames()`. In this mode, attempts to log additional
-    /// events are ignored.
-    ///
-    /// If the event contains no array of `QuicFrames` return `false`.
-    pub fn add_event_now(&mut self, event: Event) -> Result<bool> {
+    /// Writes a serializable to a JSON-SEQ record using
+    /// [std::time::Instant::now()].
+    pub fn add_event_now<E: Serialize + Eventable>(
+        &mut self, event: E,
+    ) -> Result<()> {
         let now = std::time::Instant::now();
 
         self.add_event_with_instant(event, now)
     }
 
-    /// Writes a JSON-serialized `Event` using the provided EventData and
-    /// Instant.
-    ///
-    /// Some qlog events can contain `QuicFrames`. If this is detected `true` is
-    /// returned and the streamer enters a frame-serialization mode that is only
-    /// concluded by `finish_frames()`. In this mode, attempts to log additional
-    /// events are ignored.
-    ///
-    /// If the event contains no array of `QuicFrames` return `false`.
-    pub fn add_event_with_instant(
-        &mut self, mut event: Event, now: std::time::Instant,
-    ) -> Result<bool> {
+    /// Writes a serializable to a JSON-SEQ record using the provided
+    /// [std::time::Instant].
+    pub fn add_event_with_instant<E: Serialize + Eventable>(
+        &mut self, mut event: E, now: std::time::Instant,
+    ) -> Result<()> {
         if self.state != StreamerState::Ready {
             return Err(Error::InvalidState);
         }
@@ -176,22 +166,41 @@ impl QlogStreamer {
         };
 
         let rel_time = dur.as_secs_f32() * 1000.0;
-        event.time = rel_time;
+        event.set_time(rel_time);
 
         self.add_event(event)
     }
 
-    /// Writes a JSON-serialized `Event` using the provided Instant.
-    ///
-    /// Some qlog events can contain `QuicFrames`. If this is detected `true` is
-    /// returned and the streamer enters a frame-serialization mode that is only
-    /// concluded by `finish_frames()`. In this mode, attempts to log additional
-    /// events are ignored.
-    ///
-    /// If the event contains no array of `QuicFrames` return `false`.
+    /// Writes an [Event] based on the provided [EventData] to a JSON-SEQ record
+    /// at time [std::time::Instant::now()].
+    pub fn add_event_data_now(&mut self, event_data: EventData) -> Result<()> {
+        self.add_event_data_ex_now(event_data, Default::default())
+    }
+
+    /// Writes an [Event] based on the provided [EventData] and [ExData] to a
+    /// JSON-SEQ record at time [std::time::Instant::now()].
+    pub fn add_event_data_ex_now(
+        &mut self, event_data: EventData, ex_data: ExData,
+    ) -> Result<()> {
+        let now = std::time::Instant::now();
+
+        self.add_event_data_ex_with_instant(event_data, ex_data, now)
+    }
+
+    /// Writes an [Event] based on the provided [EventData] and
+    /// [std::time::Instant] to a JSON-SEQ record.
     pub fn add_event_data_with_instant(
         &mut self, event_data: EventData, now: std::time::Instant,
-    ) -> Result<bool> {
+    ) -> Result<()> {
+        self.add_event_data_ex_with_instant(event_data, Default::default(), now)
+    }
+
+    /// Writes an [Event] based on the provided [EventData], [ExData], and
+    /// [std::time::Instant] to a JSON-SEQ record.
+    pub fn add_event_data_ex_with_instant(
+        &mut self, event_data: EventData, ex_data: ExData,
+        now: std::time::Instant,
+    ) -> Result<()> {
         if self.state != StreamerState::Ready {
             return Err(Error::InvalidState);
         }
@@ -208,20 +217,15 @@ impl QlogStreamer {
         };
 
         let rel_time = dur.as_secs_f32() * 1000.0;
-        let event = Event::with_time(rel_time, event_data);
+        let event = Event::with_time_ex(rel_time, event_data, ex_data);
 
         self.add_event(event)
     }
 
-    /// Writes a JSON-serialized `Event` using the provided Event.
-    ///
-    /// Some qlog events can contain `QuicFrames`. If this is detected `true` is
-    /// returned and the streamer enters a frame-serialization mode that is only
-    /// concluded by `finish_frames()`. In this mode, attempts to log additional
-    /// events are ignored.
-    ///
-    /// If the event contains no array of `QuicFrames` return `false`.
-    pub fn add_event(&mut self, event: Event) -> Result<bool> {
+    /// Writes a JSON-SEQ-serialized [Event] using the provided [Event].
+    pub fn add_event<E: Serialize + Eventable>(
+        &mut self, event: E,
+    ) -> Result<()> {
         if self.state != StreamerState::Ready {
             return Err(Error::InvalidState);
         }
@@ -231,78 +235,9 @@ impl QlogStreamer {
         }
 
         self.writer.as_mut().write_all(b"")?;
-
-        match event.data.contains_quic_frames() {
-            // If the event contains frames, we need to remove the closing JSON
-            // delimiters before writing out. These will be later restored by
-            // `finish_frames()`.
-            Some(frames_count) => {
-                match serde_json::to_string(&event) {
-                    Ok(mut ev_out) => {
-                        ev_out.truncate(ev_out.len() - 3);
-
-                        if frames_count == 0 {
-                            self.first_frame = true;
-                        }
-
-                        self.writer.as_mut().write_all(ev_out.as_bytes())?;
-                    },
-
-                    _ => return Err(Error::Done),
-                }
-
-                self.state = StreamerState::WritingFrames;
-            },
-
-            // If the event does not contain frames, it can be written
-            // immediately in full.
-            None => {
-                serde_json::to_writer(self.writer.as_mut(), &event)
-                    .map_err(|_| Error::Done)?;
-                self.writer.as_mut().write_all(b"\n")?;
-
-                self.state = StreamerState::Ready
-            },
-        }
-
-        Ok(event.data.contains_quic_frames().is_some())
-    }
-
-    /// Writes a JSON-serialized `QuicFrame`.
-    ///
-    /// Only valid while in the frame-serialization mode.
-    pub fn add_frame(&mut self, frame: QuicFrame, last: bool) -> Result<()> {
-        if self.state != StreamerState::WritingFrames {
-            return Err(Error::InvalidState);
-        }
-
-        if !self.first_frame {
-            self.writer.as_mut().write_all(b",")?;
-        } else {
-            self.first_frame = false;
-        }
-
-        serde_json::to_writer(self.writer.as_mut(), &frame)
+        serde_json::to_writer(self.writer.as_mut(), &event)
             .map_err(|_| Error::Done)?;
-
-        if last {
-            self.finish_frames()?;
-        }
-
-        Ok(())
-    }
-
-    /// Concludes `QuicFrame` streaming serialization.
-    ///
-    /// Only valid while in the frame-serialization mode.
-    pub fn finish_frames(&mut self) -> Result<()> {
-        if self.state != StreamerState::WritingFrames {
-            return Err(Error::InvalidState);
-        }
-
-        self.writer.as_mut().write_all(b"]}}\n")?;
-
-        self.state = StreamerState::Ready;
+        self.writer.as_mut().write_all(b"\n")?;
 
         Ok(())
     }
@@ -312,15 +247,31 @@ impl QlogStreamer {
     pub fn writer(&self) -> &Box<dyn std::io::Write + Send + Sync> {
         &self.writer
     }
+
+    pub fn start_time(&self) -> std::time::Instant {
+        self.start_time
+    }
+}
+
+impl Drop for QlogStreamer {
+    fn drop(&mut self) {
+        let _ = self.finish_log();
+    }
 }
 
 #[cfg(test)]
 mod tests {
+    use std::collections::BTreeMap;
+
     use super::*;
     use crate::events::quic;
+    use crate::events::quic::QuicFrame;
     use crate::events::RawInfo;
+    use smallvec::smallvec;
     use testing::*;
 
+    use serde_json::json;
+
     #[test]
     fn serialization_states() {
         let v: Vec<u8> = Vec::new();
@@ -345,13 +296,15 @@ mod tests {
 
         let event_data1 = EventData::PacketSent(quic::PacketSent {
             header: pkt_hdr.clone(),
-            frames: Some(vec![frame1]),
+            frames: Some(smallvec![frame1]),
             is_coalesced: None,
             retry_token: None,
             stateless_reset_token: None,
             supported_versions: None,
             raw: raw.clone(),
             datagram_id: None,
+            send_at_time: None,
+            trigger: None,
         });
 
         let ev1 = Event::with_time(0.0, event_data1);
@@ -374,26 +327,30 @@ mod tests {
 
         let event_data2 = EventData::PacketSent(quic::PacketSent {
             header: pkt_hdr.clone(),
-            frames: Some(vec![]),
+            frames: Some(smallvec![frame2]),
             is_coalesced: None,
             retry_token: None,
             stateless_reset_token: None,
             supported_versions: None,
             raw: raw.clone(),
             datagram_id: None,
+            send_at_time: None,
+            trigger: None,
         });
 
         let ev2 = Event::with_time(0.0, event_data2);
 
         let event_data3 = EventData::PacketSent(quic::PacketSent {
             header: pkt_hdr,
-            frames: Some(vec![]),
+            frames: Some(smallvec![frame3]),
             is_coalesced: None,
             retry_token: None,
             stateless_reset_token: Some("reset_token".to_string()),
             supported_versions: None,
-            raw: raw.clone(),
+            raw,
             datagram_id: None,
+            send_at_time: None,
+            trigger: None,
         });
 
         let ev3 = Event::with_time(0.0, event_data3);
@@ -411,56 +368,27 @@ mod tests {
 
         // Before the log is started all other operations should fail.
         assert!(matches!(s.add_event(ev2.clone()), Err(Error::InvalidState)));
-        assert!(matches!(
-            s.add_frame(frame2.clone(), false),
-            Err(Error::InvalidState)
-        ));
-        assert!(matches!(s.finish_frames(), Err(Error::InvalidState)));
         assert!(matches!(s.finish_log(), Err(Error::InvalidState)));
 
-        // Once a log is started, can't write frames before an event.
+        // Start log and add a simple event.
         assert!(matches!(s.start_log(), Ok(())));
-        assert!(matches!(
-            s.add_frame(frame2.clone(), false),
-            Err(Error::InvalidState)
-        ));
-        assert!(matches!(s.finish_frames(), Err(Error::InvalidState)));
+        assert!(matches!(s.add_event(ev1), Ok(())));
 
-        // Initiate log with simple event.
-        assert!(matches!(s.add_event(ev1), Ok(true)));
-        assert!(matches!(s.finish_frames(), Ok(())));
-
-        // Some events hold frames; can't write any more events until frame
-        // writing is concluded.
-        assert!(matches!(s.add_event(ev2.clone()), Ok(true)));
-        assert!(matches!(s.add_event(ev2.clone()), Err(Error::InvalidState)));
-
-        // While writing frames, can't write events.
-        assert!(matches!(s.add_frame(frame2.clone(), false), Ok(())));
-        assert!(matches!(s.add_event(ev2.clone()), Err(Error::InvalidState)));
-        assert!(matches!(s.finish_frames(), Ok(())));
-
-        // Adding an event that includes both frames and raw data should
-        // be allowed.
-        assert!(matches!(s.add_event(ev3.clone()), Ok(true)));
-        assert!(matches!(s.add_frame(frame3.clone(), false), Ok(())));
-        assert!(matches!(s.finish_frames(), Ok(())));
+        // Add some more events.
+        assert!(matches!(s.add_event(ev2), Ok(())));
+        assert!(matches!(s.add_event(ev3.clone()), Ok(())));
 
         // Adding an event with an external time should work too.
         // For tests, it will resolve to 0 but we care about proving the API
         // here, not timing specifics.
         let now = std::time::Instant::now();
 
-        assert!(matches!(
-            s.add_event_with_instant(ev3.clone(), now),
-            Ok(true)
-        ));
-        assert!(matches!(s.add_frame(frame3.clone(), false), Ok(())));
-        assert!(matches!(s.finish_frames(), Ok(())));
+        assert!(matches!(s.add_event_with_instant(ev3, now), Ok(())));
 
         assert!(matches!(s.finish_log(), Ok(())));
 
         let r = s.writer();
+        #[allow(clippy::borrowed_box)]
         let w: &Box<std::io::Cursor<Vec<u8>>> = unsafe { std::mem::transmute(r) };
 
         let log_string = r#"{"qlog_version":"version","qlog_format":"JSON-SEQ","title":"title","description":"description","trace":{"vantage_point":{"type":"server"},"title":"Quiche qlog trace","description":"Quiche qlog trace description","configuration":{"time_offset":0.0}}}
@@ -474,4 +402,143 @@ mod tests {
 
         assert_eq!(log_string, written_string);
     }
+
+    #[test]
+    fn stream_json_event() {
+        let data = json!({"foo": "Bar", "hello": 123});
+        let ev = events::JsonEvent {
+            time: 0.0,
+            importance: events::EventImportance::Core,
+            name: "jsonevent:sample".into(),
+            data,
+        };
+
+        let v: Vec<u8> = Vec::new();
+        let buff = std::io::Cursor::new(v);
+        let writer = Box::new(buff);
+
+        let trace = make_trace_seq();
+
+        let mut s = streamer::QlogStreamer::new(
+            "version".to_string(),
+            Some("title".to_string()),
+            Some("description".to_string()),
+            None,
+            std::time::Instant::now(),
+            trace,
+            EventImportance::Base,
+            writer,
+        );
+
+        assert!(matches!(s.start_log(), Ok(())));
+        assert!(matches!(s.add_event(ev), Ok(())));
+        assert!(matches!(s.finish_log(), Ok(())));
+
+        let r = s.writer();
+        #[allow(clippy::borrowed_box)]
+        let w: &Box<std::io::Cursor<Vec<u8>>> = unsafe { std::mem::transmute(r) };
+
+        let log_string = r#"{"qlog_version":"version","qlog_format":"JSON-SEQ","title":"title","description":"description","trace":{"vantage_point":{"type":"server"},"title":"Quiche qlog trace","description":"Quiche qlog trace description","configuration":{"time_offset":0.0}}}
+{"time":0.0,"name":"jsonevent:sample","data":{"foo":"Bar","hello":123}}
+"#;
+
+        let written_string = std::str::from_utf8(w.as_ref().get_ref()).unwrap();
+
+        assert_eq!(log_string, written_string);
+    }
+
+    #[test]
+    fn stream_data_ex() {
+        let v: Vec<u8> = Vec::new();
+        let buff = std::io::Cursor::new(v);
+        let writer = Box::new(buff);
+
+        let trace = make_trace_seq();
+        let pkt_hdr = make_pkt_hdr(quic::PacketType::Handshake);
+        let raw = Some(RawInfo {
+            length: Some(1251),
+            payload_length: Some(1224),
+            data: None,
+        });
+
+        let frame1 = QuicFrame::Stream {
+            stream_id: 40,
+            offset: 40,
+            length: 400,
+            fin: Some(true),
+            raw: None,
+        };
+
+        let event_data1 = EventData::PacketSent(quic::PacketSent {
+            header: pkt_hdr.clone(),
+            frames: Some(smallvec![frame1]),
+            is_coalesced: None,
+            retry_token: None,
+            stateless_reset_token: None,
+            supported_versions: None,
+            raw: raw.clone(),
+            datagram_id: None,
+            send_at_time: None,
+            trigger: None,
+        });
+        let j1 = json!({"foo": "Bar", "hello": 123});
+        let j2 = json!({"baz": [1,2,3,4]});
+        let mut ex_data = BTreeMap::new();
+        ex_data.insert("first".to_string(), j1);
+        ex_data.insert("second".to_string(), j2);
+
+        let ev1 = Event::with_time_ex(0.0, event_data1, ex_data);
+
+        let frame2 = QuicFrame::Stream {
+            stream_id: 1,
+            offset: 0,
+            length: 100,
+            fin: Some(true),
+            raw: None,
+        };
+
+        let event_data2 = EventData::PacketSent(quic::PacketSent {
+            header: pkt_hdr.clone(),
+            frames: Some(smallvec![frame2]),
+            is_coalesced: None,
+            retry_token: None,
+            stateless_reset_token: None,
+            supported_versions: None,
+            raw: raw.clone(),
+            datagram_id: None,
+            send_at_time: None,
+            trigger: None,
+        });
+
+        let ev2 = Event::with_time(0.0, event_data2);
+
+        let mut s = streamer::QlogStreamer::new(
+            "version".to_string(),
+            Some("title".to_string()),
+            Some("description".to_string()),
+            None,
+            std::time::Instant::now(),
+            trace,
+            EventImportance::Base,
+            writer,
+        );
+
+        assert!(matches!(s.start_log(), Ok(())));
+        assert!(matches!(s.add_event(ev1), Ok(())));
+        assert!(matches!(s.add_event(ev2), Ok(())));
+        assert!(matches!(s.finish_log(), Ok(())));
+
+        let r = s.writer();
+        #[allow(clippy::borrowed_box)]
+        let w: &Box<std::io::Cursor<Vec<u8>>> = unsafe { std::mem::transmute(r) };
+
+        let log_string = r#"{"qlog_version":"version","qlog_format":"JSON-SEQ","title":"title","description":"description","trace":{"vantage_point":{"type":"server"},"title":"Quiche qlog trace","description":"Quiche qlog trace description","configuration":{"time_offset":0.0}}}
+{"time":0.0,"name":"transport:packet_sent","data":{"header":{"packet_type":"handshake","packet_number":0,"version":"1","scil":8,"dcil":8,"scid":"7e37e4dcc6682da8","dcid":"36ce104eee50101c"},"raw":{"length":1251,"payload_length":1224},"frames":[{"frame_type":"stream","stream_id":40,"offset":40,"length":400,"fin":true}]},"first":{"foo":"Bar","hello":123},"second":{"baz":[1,2,3,4]}}
+{"time":0.0,"name":"transport:packet_sent","data":{"header":{"packet_type":"handshake","packet_number":0,"version":"1","scil":8,"dcil":8,"scid":"7e37e4dcc6682da8","dcid":"36ce104eee50101c"},"raw":{"length":1251,"payload_length":1224},"frames":[{"frame_type":"stream","stream_id":1,"offset":0,"length":100,"fin":true}]}}
+"#;
+
+        let written_string = std::str::from_utf8(w.as_ref().get_ref()).unwrap();
+
+        assert_eq!(log_string, written_string);
+    }
 }
